/*
 * Project   : SIMUNO
 * Package   : simuno.imunology.objects
 * File      : Aminoacid.java
 * Created at: 04/07/2007 - 19:49:28
 */
package simuno.imunology.objects;

import simuno.AminoacidTable;

/**
 * Class representing an aminoacid. * simuno.imunology.objects.Aminoacid
 * @author Rodrigo Janurio da Silva
 * @version 1.0.0
 */
public final class Aminoacid
{
    /** The aminoacid id. */
    private long id;
    
    /** The aminoacid type */
    private String threeLetterName;
    
    /** The aminoacid name */
    private String name;
    
    /** The aminoacid formula. */
    private String formula;
    
    /** The hydropathy index. Bigger it is, more hydrophobic is the aminoacid */
    private double hydropathIndex;
    
    /** The charge of the aminoacid. Can be -1 (negative), 1 (positive) or 0 (no charge) */
    private int charge;
	
    /**
     * Creates a new instance of aminoacid passing a type as parameter.
     * @param type The type to be set to the aminoacid.
     * @throws IllegalArgumentException if the type is null.
     */
    public Aminoacid( final String threeLetterName
                    , final String name
                    , final String formula
                    , final double hydropathIndex
                    , final int    charge)
        throws IllegalArgumentException
    {
        this.validateParameters( threeLetterName
                               , name
                               , formula
                               , charge );
        this.id              = IdGenerator.generateId( this.getClass() );
        this.name            = name;
        this.charge          = charge;
        this.formula         = formula;
        this.hydropathIndex  = hydropathIndex;
        this.threeLetterName = threeLetterName;
    }
    
    /**
     * Builds a new instance of Aminoacid based only on the three letter name passed as parameter.
     * It no aminoacid exists with that letters, throw an exception.
     * @param threeLetterName The three letter name of the aminoacid
     * @throws IllegalArgumentException if the three letters are not valid.
     */
    public Aminoacid ( final String threeLetterName )
        throws IllegalArgumentException
    {
        Aminoacid aminoacid = AminoacidTable.searchByThreeLetterName( threeLetterName );
        
        if ( null != aminoacid )
        {
            // Just copy the aminoacid found
            this.id = IdGenerator.generateId( this.getClass() );
            this.copy( aminoacid );
        }
        else
        {
            // Ups.. Invalid letters. 
            throw new IllegalArgumentException( "The letters passed " + threeLetterName + 
            		" do not represent a valid aminoacid!" );
        }
    }
    
    /**
     * Copies the passed aminoacid as parameter into this.
     * @param aminoacid The aminoacid to be copied
     * @throws NullPointerException If the aminoacid passed is null.
     */
    public void copy ( Aminoacid aminoacid )
        throws NullPointerException
    {
        if ( null == aminoacid )
        {
            throw new NullPointerException( "The aminoacid can't be null!" );
        }
        
        this.name            = aminoacid.name;
        this.charge          = aminoacid.charge;
        this.formula         = aminoacid.formula;
        this.hydropathIndex  = aminoacid.hydropathIndex;
        this.threeLetterName = aminoacid.threeLetterName;
    }

    /**
     * This method validates a type. The type can be anything but empty or null.
     * @param type The type to be set..
     * @throws IllegalArgumentException If the type is empty or null..
     */
    private void validateParameters( final String threeLetterName
                                   , final String name
                                   , final String formula
                                   , final int    charge )
        throws IllegalArgumentException
    {
        // The type can't be null.
        if ( ( null == threeLetterName ) ||
             ( threeLetterName.trim().equals( "" ) ) ||
             ( threeLetterName.length() != 3 ) )
        {
            throw new IllegalArgumentException( "The threeLetterName can't neither null nor empty!" +
            		" It also can have only three letters." );
        }
        
        // Verify for existent aminoacid with that letters.
        if ( null == AminoacidTable.searchByThreeLetterName( threeLetterName ) )
        {
            throw new IllegalArgumentException( "There is no valid aminoacids with the letters " +
            		threeLetterName + "." );
        }

        // The name can't be null.
        if ( ( null == name ) ||
            ( name.trim().equals( "" ) ) )
        {
            throw new IllegalArgumentException( "The name can't neither null nor empty!" );
        }

        // The formula can't be null.
        if ( ( null == formula ) ||
            ( formula.trim().equals( "" ) ) )
        {
            throw new IllegalArgumentException( "The formula can't neither null nor empty!" );
        }
        
        // Thc charge can be -1, 0 or 1
        if ( ( -1 > charge ) ||
             ( 1  < charge ) )
        {
            throw new IllegalArgumentException( "The charge can only be -1, 0 or 1!" );
        }
    }
    
    /**
     * Returns the type to the caller.
     * @return The aminoacid type.
     */
    public final String getThreeLetterName()
    {
        return this.threeLetterName;
    }
    
    /**
     * Returns the name to the caller.
     * @return The aminoacid name.
     */
    public final String getName()
    {
        return this.name;
    }

    /**
     * Returns the formula to the caller.
     * @return The aminoacid formula.
     */
    public final String getFormula()
    {
        return this.formula;
    }

    /**
     * Returns to the caller the aminoacid id.
     * @return The aminoacid id..
     */
    public final long getId()
    {
        return this.id;
    }
    
    /**
     * Returns to the caller the aminoacid hydropathy index;
     * @return The aminoacid hydropathy index
     */
    public final double getHydropathyIndex()
    {
        return this.hydropathIndex;
    }

    /**
     * Returns to the caller the aminoacid charge;
     * @return The aminoacid charge.
     */
    public final int getCharge()
    {
        return this.charge;
    }

    /*
     * (non-Javadoc)
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals( Object anObject )
    {
        return this.threeLetterName.equals( ( ( Aminoacid ) anObject ).getThreeLetterName() );
    }
}
