/*
 * Project   : SIMUNO
 * Package   : simuno.imunology.objects
 * File      : Epitop.java
 * Created at: 09/07/2007 - 12:31:05
 */
package simuno.imunology.objects;

import java.util.ArrayList;

/**
 * Class representing an epitop. An epitop is an aminoacid sequence. It's like a small protein that
 * have only between 5 and 17 aminoacids. It comes from the protein process by the APC cell. It's
 * what the APC shows for the T4 lymphocyte to recognize. 
 * simuno.imunology.objects.Epitop
 * @author Rodrigo Janurio da Silva.
 * @version 1.0.0
 */
public final class Epitop
    extends AminoacidSequence
{
    /** The epitop identificator. */
    private long id;
    
    /** The minimum number of the aminoacids in sequence. */
    public static final short MIN_NUM_AMINOACID = 5;

    /** The maximum number of aminoacids in the sequence.. */
    public static final short NUM_MAX_AMINOACID = 17;
    
    /**
     * Creates a new instance of epitop. Its id is generated here.
     */
    public Epitop()
    {
        super();
        this.id = IdGenerator.generateId( this.getClass() );
    }
    
    /**
     * Creates a new instance of epitop. It generates the id and also sets the aminoacid sequence
     * as the one passed.
     * @param aminoacidSeq Aminoacid sequence to be set.
     * @throws IllegalArgumentException If the sequence passed is null. 
     */
    public Epitop( final ArrayList<Aminoacid> seqAminoacidos )
        throws IllegalArgumentException
    {
        super( seqAminoacidos );
        
        if ( ( seqAminoacidos.size() < Epitop.MIN_NUM_AMINOACID ) ||
             ( seqAminoacidos.size() > Epitop.NUM_MAX_AMINOACID ) )
        {
            throw new IllegalArgumentException( "The epitop aminoacid sequence should have between " +
            		"5 and 17 aminoacids." );
        }
    }

    /**
     * Returns to the user the epitop id.
     * @return The user the protein id.
     */
    public final long getId()
    {
        return id;
    }
    
    /*
     * (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
    	String ret = "";
    	
    	for( Aminoacid a : this.aminoacids )
    	{
    		ret += a.getThreeLetterName() + "|";
    	}
    	
    	return ret;
    }
}
