/*
 * Project   : SIMUNO
 * Package   : simuno.imunology.objects
 * File      : APC.java
 * Created at: 27/07/2007 - 23:57:30
 */
package simuno.imunology.objects.cells;

import java.util.ArrayList;
import java.util.Observer;

import simuno.RandomGenerator;
import simuno.Sleeper;
import simuno.imunology.Enviroment;
import simuno.imunology.objects.Aminoacid;
import simuno.imunology.objects.Antigen;
import simuno.imunology.objects.Epitop;
import simuno.imunology.objects.IdGenerator;
import simuno.imunology.objects.Message;
import simuno.imunology.objects.conectors.MHC;
import simuno.statistics.StatisticValue;
import simuno.statistics.Statistics;
import simuno.statistics.StatisticsType;

/**
 * This class represents an APC. This cell is in charge of eating the antigens, process them into
 * epitop and present them to the lymphocyte through the MHC.
 * simuno.imunology.objects.APC
 * @author Rodrigo Janurio da Silva.
 * @version 1.0.0
 */
public final class APC
    extends Cell
{
    /** APC identificator. */
    private long id;
    
    /** This flag tells us if the APC is activated by PAMPs or not. */
    private boolean activated;
    
    /** This is the eaten antigen. */
    private Antigen antigen;
    
    /** The processed epitop. */
    private Epitop epitop;
    
    private String name;
    
    private boolean connected;

    /**
     * Creates a new APC instance. An observer is passed for this.
     * @param o The observer for this APC
     */
    public APC( Observer o )
    {
        super( o );
        this.id        = IdGenerator.generateId( this.getClass() );
        this.name      = "APC " + this.id;
        this.activated = false;
        this.epitop    = null;
        this.connected = false;
        this.conectors.add( new MHC() );
    }

    /**
     * Returns the APC status, if it is activated or not.
     * @return <b>true</b> if the cell is activated. <b>false</b> otherwise.
     */
    public final boolean isActivated()
    {
        return this.activated;
    }
    
    public final boolean isConnected()
    {
    	return this.connected;
    }
    
    /**
     * Sets the APC status.
     * @param activated The status to be set.
     */
    public final void setActivated( boolean activated )
    {
        this.activated = activated;
    }

    public final void setConnected( boolean connected )
    {
    	this.connected = connected;
    }

    /**
     * Returns the APC id.
     * @return The APC id.
     */
    public final long getId()
    {
        return this.id;
    }
    
    public final String getName()
    {
    	return this.name;
    }

    /**
     * Method in charge to simulate the antigen eat.
     * @param antigen The antigen being eat.
     * @throws NullPointerException if the parameter is null.
     */
    public final void eatAntigen( final Antigen antigen )
        throws NullPointerException
    {
        if ( null != antigen )
        {
        	//Sleeper.sleepThread( 500 );
	        this.antigen   = antigen;
	        this.activated = true;
	        Enviroment.getInstance().removeAntigen( antigen );
	        this.notifyObservers( new StatisticValue( StatisticsType.APC_NUM_ANTIGEN_EATEN
	                                                , new Integer( 1 ) ) );
        }
    }
    
    /**
     * Method in charge to process the Antigen in epitops. The first aminoacid is achieved
     * randomically.
     */
    public final void processAntigenInEpitop()
    {
        int                  numAminoacidInEpitop = 0;
        int                  initialAminoIdx      = -1;
        ArrayList<Aminoacid> outAminoacids        = new ArrayList<Aminoacid>();
        ArrayList<Aminoacid> inAminoacids         = this.antigen.getAminoacids();
        
        // Setting the number of aminoacid that is going to be extracted.
        numAminoacidInEpitop = RandomGenerator.newInt() % MHC.MhcSize;
        numAminoacidInEpitop = numAminoacidInEpitop < 5 ? 5 : numAminoacidInEpitop;
        
        // Setting the initial aminoacid index.
        initialAminoIdx      = RandomGenerator.newInt() % inAminoacids.size();
        initialAminoIdx      = initialAminoIdx + numAminoacidInEpitop >= inAminoacids.size() ?
                                   inAminoacids.size() - numAminoacidInEpitop - 1 :
                                   initialAminoIdx;  

        // Getting the aminoacids for creating the epitop.
        for( int i = 0; i < numAminoacidInEpitop; i++ )
        {
            outAminoacids.add( inAminoacids.get( i + initialAminoIdx ) );
            //Sleeper.sleepThread( 75 );
        }
        
        this.epitop = new Epitop( outAminoacids );
        
        // Cleanning variables.
        outAminoacids = null;
        inAminoacids  = null;
    }
    
    /**
     * Puts the processed epitop in the MHC for presenting.
     */
    public void presentEpitop()
    {
        this.getMHC().setEpitop( this.epitop );
        this.notifyObservers( new StatisticValue( StatisticsType.APC_NUM_EPITOP_PRESENTED
                                                , new Integer( 1 ) ) );
    }
    
    /**
     * An APC returns only answer the CD28 with a B7 when it's activated by pamps.
     * @return <b>true</b> if answered. <b>false</b> otherwise.
     */
    public boolean answerCD28WithB7()
    {
        this.notifyObservers( new StatisticValue( StatisticsType.APC_NUM_B7_ANSWERS
                                                , new Integer( 1 ) ) );
        return activated;
    }

    /**
     * This method processes a message received from a cell. The messages accepted are:
     *     - CD28: When a lymphocyte sends a signal like that to the APC, it's answered with a B7
     *             costimulation.
     * @param msg The message received.
     * @param cel The cell that sents the message.
     * @throws NullPointerException if some of the parameters are null.
     */
    @Override
    public void receiveMessage( Message msg
                              , Cell    cel )
    {
        super.receiveMessage( msg, cel );
        
        switch( msg )
        {
            case CD28:
            {
            	//Sleeper.sleepThread( 500 );
            	System.out.println( this.name + ": Recebi mensagem CD28..." );
                if ( this.activated )
                {
                    if ( cel instanceof T4 )
                    {
                        T4 t4 = ( T4 ) cel;
                        System.out.println( this.name + ": Enviando mensagem B7..." );
                        try
                        {
                            Thread.sleep( 100 );
                        }
                        catch ( InterruptedException e )
                        {
                            e.printStackTrace();
                        }
                        t4.receiveMessage( Message.B7, this );
                        this.notifyObservers( new StatisticValue( StatisticsType.APC_NUM_B7_ANSWERS
                                                                , new Integer( 1 ) ) );
                    }
                }
            }
            break;
        }
    }

    /**
     * This method implements the APC behavior. It searches in the enviroment if there is a antigen
     * inside. When it is found, it is eaten, processed in epitops and then presented on the MHC.
     */
    @Override
    public void run()
    {
        while ( !this.stop )
        {
            Enviroment env = Enviroment.getInstance();
            
            while ( !env.hasAntigens() && !this.stop )
            {
                Sleeper.sleepThread( 50 );
            	if ( this.activated )
            	{
            		this.discardEpitop();
            		( ( Statistics ) this.observers.get( 0 ) ).printStatistics( System.err );
            	}
            	
                System.out.println( this.name + ": Procurando antgeno..." );
            }
            
            if ( !this.stop )
            {
                // Only eat an antigen if its not activated.
                if ( !this.activated )
                {
                	Antigen antigen = null;
                	
                	// I'm eating this antigen, so it's mine
                	synchronized ( env.getAntigens() )
                	{
                    	int index = RandomGenerator.newInt() % env.getAntigens().size();
    
                    	// This antigen is mine.
                		antigen = env.getAntigens().get( index );
                        System.out.println( this.name + ": Fagocitando antgeno..." );
                        this.eatAntigen( antigen );
                        System.out.println( this.name + ": Processando antgeno em eptopo..." );
                        this.processAntigenInEpitop();
                        System.out.println( this.name + ": Apresentando eptopo..." );
                        this.presentEpitop();
                        Sleeper.sleepThread( 10000 );
    				}
                }
            }
        }
    }

    /**
     * Simulates the APC reproduction.
     */
    @Override
    public void reproduce()
    {
    }
    
    public void discardEpitop()
    {
    	System.out.println( this.name + ": Resposta imunolgica terminada. Discartando eptopo..." );
    	( ( MHC ) this.conectors.get( 0 ) ).setEpitop( null );
    	this.activated = false;
    }
    
    public MHC getMHC()
    {
    	return ( MHC ) this.conectors.get( 0 );
    }
}
