/*
 * Project   : SIMUNO
 * Package   : simuno.imunology.objects.cells
 * File      : T4.java
 * Created at: 10/08/2007 - 08:46:41
 */
package simuno.imunology.objects.cells;

import java.util.Observer;

import simuno.RandomGenerator;
import simuno.Sleeper;
import simuno.imunology.Enviroment;
import simuno.imunology.objects.Antigen;
import simuno.imunology.objects.Epitop;
import simuno.imunology.objects.IdGenerator;
import simuno.imunology.objects.Message;
import simuno.statistics.StatisticValue;
import simuno.statistics.StatisticsType;

/**
 * This class represents a lymphocyte T4. It's already specialized because it has already passed for
 * the positive selection and in this stage it recognized a MHC II. It has already passed by the
 * stage of negative selection, when the lymphocytes are tested againt the auto-antigen. Those ones
 * who recognize it are thrown away and enter in apoptose.
 * simuno.imunology.objects.cells.T4
 * @author Rodrigo Janurio da Silva
 * @author Marcos Tiago Garcia.
 * @version 1.0.0
 */
public class T4
    extends Lymphocyte
{
    /** The T4 lymphocyte id. */
    private long id;
    
    /** This flag indicates when a T4 lymphocyte receives successfully a B& answer from the APC. */
    private boolean b7Ok;
    
    /** This holds the recognized epitop. */
    private Epitop recognizedEpitop;
    
    /** This flag indicates if a lymphocyte is a memory lymphocite or not. */
    private boolean memory;
    
    /** The lymphocyte name. (Only for make prints easy) */
    private String name;
    
    /**
     * Creates a new instance of T4.
     * @param o The observer for this Lymphocyte.
     */
    public T4( Observer o )
    {
        super( o );
        this.id               = IdGenerator.generateId( this.getClass() );
        this.name             = "T4 " + this.id;
        this.b7Ok             = false;
        this.memory           = false;
        this.recognizedEpitop = null;
    }
    
    /**
     * Returns the T4 identificator to the caller.
     * @return
     */
    public final long getId()
    {
        return this.id;
    }
    
    /**
     * This method implements the behavior of a T4 lymphocite. A T4 cell searches for APC's and when
     * they (APC) are activated by PAMP's or other strange substance they can start a comunication.
     * If the comunication succeds then the lymphocite starts the epitp reconition stage.
     */
    @Override
    public void run()
    {
        Enviroment env   = Enviroment.getInstance();
        
        while ( !this.stop )
        {
        	Sleeper.sleepThread( 50 );
        	int index = -1;
        	
        	if ( null == this.recognizedEpitop )
        	{
	        	// Getting randomically a cell and verifying if it's an APC.
	        	index      = RandomGenerator.newInt() % env.getCells().size();
	        	Cell cell  = env.getCells().get( index );
	                
	            if ( cell instanceof APC )
	            {
	                APC apc = null;
	                
	                // Found it! Now, I'm talking with this APC so this is mine.
	            	synchronized ( cell )
	            	{
						apc = ( APC ) cell;
						if ( !apc.isConnected() )
						{
							// Now I'm connected with the APC. No one can connect with it now.
							apc.setConnected( true );
		                    System.out.println( this.name + ": Enviando mensagem CD28 para " +
		                    		apc.getName() + "..." );
		                    this.b7Ok = false;
		                    apc.receiveMessage( Message.CD28, this );
		                    this.notifyObservers( new StatisticValue( 
		                    		                  StatisticsType.T4_NUM_CD28_SENT
	                                                , new Integer( 1 ) ) );
		                    if ( this.b7Ok )
		                    {
		                        Epitop e = apc.getMHC().getEpitop();
		                    	if ( null != e && this.recognizeEpitop( e ) )
			                    {
			                    	System.out.println( this.name + ": Recebi B7 de " +
			                    			apc.getName() + "..." );
			                    	System.out.println( this.name + ": Epitopo de " + apc.getName() +
			                    			" reconhecido -> " + e + "..." );
			                    }
			                    else
			                    {
			                    	System.out.println( this.name + ": Epitopo de " + apc.getName() +
			                    			" no reconhecido -> " + e + "..." );
			                    }
		                    }
		                    	 
	                    	apc.setConnected( false );
						}
					}
	            }
        	}
        	else 
        	{
            	System.out.println( this.name + ": Procurando Antgenos..." );
            	//Sleeper.sleepThread( 500 );
        	
            	if ( env.hasAntigens() )
            	{
            		Antigen antigen = null;
            		
            		synchronized ( env.getAntigens() )
            		{
                		// Getting randomically an antigen.
                		if ( env.getAntigens().size() > 0 )
                		{
                			index   = RandomGenerator.newInt() % env.getAntigens().size();
                			antigen = env.getAntigens().get( index );
                		}
                		
            			// This antigen is mine!
            			System.out.println( this.name + ": Antgeno Encontrado. Iniciando " +
            					"reconhecimento..." );
            			if ( this.recognizeAntigen( antigen ) )
            			{
            				System.out.println( this.name + ": Antgeno reconhecido. " +
            						"Destruindo..." );
            				this.killAntigen( antigen );
            			}
            			else
            			{
            				System.out.println( this.name + ": Antgeno no reconhecido..." );
            			}
					}
            	}
            	else
            	{
            		this.b7Ok             = false;
            		this.recognizedEpitop = null;
            	}
            }
        }
    }

    /**
     * This method is in charge to treat the received messages from other cells. The messages
     * treated are the following:
     *     - B7: A co-stimulation from an APC cell after receiving a CD28 stimulation.
     * @param msg The message received.
     * @param cel The cel that sent the message.
     * @throws NullPointerException if some of the parameters are null.
     */
    @Override
    public void receiveMessage( Message msg
                              , Cell    cel )
        throws NullPointerException
    {
        super.receiveMessage( msg, cel );
        
        switch( msg )
        {
            case B7:
            {
                if ( cel instanceof APC )
                {
                    this.b7Ok = true;
                }
            }
            break;
        }
    }
    
    /**
     * This method is in charge of simulating an epitop rocognition. In a first moment we are going
     * to work with yes / no answer, randomically.
     * @param epitop The epitop to be recognized.
     * @return <b>true</b> if recognized the epitop. <b>false</b> otherwise.
     * @throws NullPointerException if the parameter is null.
     */
    @Override
    public boolean recognizeEpitop( final Epitop epitop )
        throws NullPointerException
    {
        boolean recognized = false;
        
        //Sleeper.sleepThread( 1500 );
        if ( ( null != epitop ) && 
        	 ( RandomGenerator.newInt() % 10 >= 7 ) )
        {
            recognized = true;
            this.recognizedEpitop = epitop;
            this.notifyObservers( new StatisticValue( StatisticsType.T4_NUM_EPITOP_RECOGNIZED
                                                    , new Integer( 1 ) ) );
        }
        else
        {
            this.notifyObservers( new StatisticValue( 
            		                  StatisticsType.T4_NUM_THAT_DID_NOT_RECOGNIZED_ANYTHING
                                    , new Integer( 1 ) ) );
        }

        return recognized;
    }
    
    /**
     * This method tries to recognize a Antigen based on the epitop recognized before on APC.
     * @param antigen The antigen to be recognized.
     * @return <b>true</b> if the antigen is recognized. <b>false</b> otherwise.
     */
    @Override
    public boolean recognizeAntigen( final Antigen antigen )
    {
        boolean recognized = false;
        
        if ( null != antigen )
        {
	        // Searching for the epitop in the antigen
	        for ( int i = 0; null != antigen && i < antigen.getAminoacids().size(); i++ )
	        {
	            int j = 0;
	            
	            // For each aminoacid in the epitop we see if it matches in the antigen
	            for ( ; j < this.recognizedEpitop.getAminoacids().size(); j++ )
	            {
	            	//Sleeper.sleepThread( 75 );
	                if ( !antigen.getAminoacids().get( i + j ).equals( 
	                        this.recognizedEpitop.getAminoacids().get( j ) ) )
	                {
	                    // Do not match. Let's get outta here.
	                    break;
	                }
	            }
	            
	            if ( j >= this.recognizedEpitop.getAminoacids().size() )
	            {
	                // Went throught the final. Let's return.
	                recognized = true;
	                break;
	            }
	        }
	        
	        if ( recognized )
	        {
	            this.notifyObservers( new StatisticValue( 
	            		                  StatisticsType.T4_NUM_ANTIGEN_RECOGNIZED
	                                    , new Integer( 1 ) ) );
	        }
        }
        
        return recognized;
    }

    /**
     * This mehtod simulates a lymphocite reproduction.
     */
    @Override
    public void reproduce()
    {
        Enviroment env = Enviroment.getInstance();
        T4         t4  = new T4( this.observers.get( 0 ) );
        
        if ( RandomGenerator.newInt() % 100 >= 75 )
        {
            t4.memory = true;
        }
        
        env.getCells().add( t4 );
    }

    /**
     * @return the memory
     */
    public final boolean isMemory()
    {
        return this.memory;
    }
}
