/*
 * Project   : SIMUNO
 * Package   : simuno.imunology.objects
 * File      : MHC.java
 * Created at: 09/07/2007 - 22:17:24
 */
package simuno.imunology.objects.conectors;

import java.util.ArrayList;
import java.util.Iterator;

import simuno.RandomGenerator;
import simuno.imunology.objects.Epitop;
import simuno.imunology.objects.Gene;
import simuno.imunology.objects.IdGenerator;

/**
 * This class represents a MHC. MHC is a part on the APC where the antigen epitop are shown for
 * lymphocyte recognition.
 * simuno.imunology.objects.MHC
 * @author Rodrigo Janurio da Silva
 * @version 1.0.0
 */
public final class MHC
    extends Conector
{
    /** The MHC id. */
    private long id;
    
    /** The MHC forming genes (A, B, C, DP, DQ, DR). */
    private ArrayList<Gene> genes;
    
    /** Epitop being presented on MHC. */
    private Epitop epitop;
    
    /** The number of aminoacids that are handled by this MHC. */
    public static int MhcSize = 0;
    
    /** Number of genes contained on MHC. */
    public static final short NUM_GENES = 6;
    
    // static initialization.
    static
    {
        MHC.MhcSize = RandomGenerator.newInt() % 13 + 5;
    }
    
    /**
     * Builds a new instance of MHC. Here the id is generated and the the genes are initialized.
     */
    public MHC()
    {
        this.id     = IdGenerator.generateId( this.getClass() );
        this.genes  = new ArrayList<Gene>( MHC.NUM_GENES );
        this.epitop = null;
        
        // Putting the gens inside the MHC.
        this.genes.add( new Gene( "A" , true ) );
        this.genes.add( new Gene( "B" , true ) );
        this.genes.add( new Gene( "C" , true ) );
        this.genes.add( new Gene( "DP", !true ) );
        this.genes.add( new Gene( "DQ", !true ) );
        this.genes.add( new Gene( "DR", !true ) );
    }
    
    /**
     * Returns the MHC identifier.
     * @return The MHC id.
     */
    public final long getId()
    {
        return this.id;
    }
    
    /**
     * Returns all the 6 MHC genes.
     * @return The MHC gene array.
     */
    public final ArrayList<Gene> getGenes()
    {
        return this.genes;
    }
    
    /**
     * Returns to the caller if this MHC is one of type I.
     * @return <B>true</B> if the MHC is type I. <B>false</B> otherwise.
     */
    public final boolean isMhcTypeI()
    {
        return false;
    }
    
    /**
     * Returns to the caller only the active gens on the MHC.
     * @return An array containing all the active gens on the MHC.
     */
    public final ArrayList<Gene> getActiveGens()
    {
        ArrayList<Gene> activeGens = new ArrayList<Gene>();
        Iterator<Gene>  geneIt     = this.genes.iterator();
        
        while( geneIt.hasNext() )
        {
            Gene gene = geneIt.next();
            if ( true == gene.isActive() )
            {
                activeGens.add( gene );
            }
        }
        
        return activeGens;
    }

    /**
     * Method that puts the epitop on the MHC. After setted the epitop can be shown.
     * @param epitop Epitop to be put.
     */
    public void setEpitop( final Epitop epitop )
    {
        this.epitop = epitop;
    }
    
    /**
     * Returns the epitop being presented.
     * @return The epitop being presented..
     */
    public Epitop getEpitop()
    {
        return this.epitop;
    }
}
